class NoiseProtocolNameError(Exception):
    pass


class NoisePSKError(Exception):
    pass


class NoiseValueError(Exception):
    pass


class NoiseHandshakeError(Exception):
    pass


class NoiseInvalidMessage(Exception):
    pass


class NoiseMaxNonceError(Exception):
    pass


class NoiseValidationError(Exception):
    pass
