from noise.backends.default import DefaultNoiseBackend


class ExperimentalNoiseBackend(DefaultNoiseBackend):
    """
    Contains all the default crypto methods, but also methods not directly endorsed by Noise Protocol specification
    """
    pass
